@@ -3,7 +3,8 @@
    Data Link Provider Interface (DLPI) network interface code. */
 
 /*
- * Copyright (c) 2004,2007,2009 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2009-2011 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004,2007 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -981,7 +982,7 @@ static int dlpibindack (fd, bufp)
 	
 	dlp = (union DL_primitives *)ctl.buf;
 	
-	if (!expected (DL_BIND_ACK, dlp, flags) < 0) {
+	if (expected (DL_BIND_ACK, dlp, flags) == -1) {
 		return -1;
 	}
 	
@@ -1015,7 +1016,7 @@ static int dlpiokack (fd, bufp)
 	
 	dlp = (union DL_primitives *)ctl.buf;
 	
-	if (!expected (DL_OK_ACK, dlp, flags) < 0) {
+	if (expected (DL_OK_ACK, dlp, flags) == -1) {
 		return -1;
 	}
 	
@@ -1049,7 +1050,7 @@ static int dlpiinfoack (fd, bufp)
 	
 	dlp = (union DL_primitives *) ctl.buf;
 	
-	if (!expected (DL_INFO_ACK, dlp, flags) < 0) {
+	if (expected (DL_INFO_ACK, dlp, flags) == -1) {
 		return -1;
 	}
 	
@@ -1083,7 +1084,7 @@ int dlpiphysaddrack (fd, bufp)
 
 	dlp = (union DL_primitives *)ctl.buf;
 	
-	if (!expected (DL_PHYS_ADDR_ACK, dlp, flags) < 0) {
+	if (expected (DL_PHYS_ADDR_ACK, dlp, flags) == -1) {
 		return -1;
 	}
 
@@ -1164,16 +1165,7 @@ static int dlpiunitdataind (fd, daddr, daddrlen,
 
 	result = getmsg (fd, &ctl, &data, &flags);
 
-	/*
-	 * The getmsg() manpage says:
-	 *
-	 * "On successful completion, a non-negative value is returned."
-	 *
-	 * This suggests that if MOREDATA or MORECTL are set, we error?
-	 * This seems to be safe as it never seems to happen.  Still,
-	 * set a log message, so we know if it ever starts happening.
-	 */
-	if (result != 0) {
+	if (result < 0) {
 		log_debug("dlpiunitdataind: %m");
 		return -1;
 	}
@@ -1225,15 +1217,15 @@ static int expected (prim, dlp, msgflags)
 {
 	if (msgflags != RS_HIPRI) {
 		/* Message was not M_PCPROTO */
-		return 0;
+		return -1;
 	}
 
-	if (dlp -> dl_primitive != prim) {
+	if (dlp->dl_primitive != prim) {
 		/* Incorrect/unexpected return message */
-		return 0;
+		return -1;
 	}
 	
-	return 1;
+	return 0;
 }
 
 /*
