@@ -3,7 +3,7 @@
    BSD socket interface code... */
 
 /*
- * Copyright (c) 2004-2009 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2010 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -54,6 +54,17 @@
 # endif
 #endif
 
+#if defined(DHCPv6)
+/*
+ * XXX: this is gross.  we need to go back and overhaul the API for socket
+ * handling.
+ */
+static unsigned int global_v6_socket_references = 0;
+static int global_v6_socket = -1;
+
+static void if_register_multicast(struct interface_info *info);
+#endif
+
 /*
  * If we can't bind() to a specific interface, then we can only have
  * a single socket. This variable insures that we don't try to listen
@@ -97,12 +108,18 @@ void if_reinitialize_receive (info)
 		defined (USE_SOCKET_FALLBACK)
 /* Generic interface registration routine... */
 int
-if_register_socket(struct interface_info *info, int family, int do_multicast) {
+if_register_socket(struct interface_info *info, int family,
+		   int *do_multicast)
+{
 	struct sockaddr_storage name;
 	int name_len;
 	int sock;
 	int flag;
 	int domain;
+#ifdef DHCPv6
+	struct sockaddr_in6 *addr6;
+#endif
+	struct sockaddr_in *addr;
 
 	/* INSIST((family == AF_INET) || (family == AF_INET6)); */
 
@@ -120,28 +137,41 @@ if_register_socket(struct interface_info *info, int family, int do_multicast) {
 	 * address family. 
 	 */ 
 	memset(&name, 0, sizeof(name));
+	switch (family) {
 #ifdef DHCPv6
-	if (family == AF_INET6) {
-		struct sockaddr_in6 *addr = (struct sockaddr_in6 *)&name; 
-		addr->sin6_family = AF_INET6;
-		addr->sin6_port = local_port;
-		memcpy(&addr->sin6_addr,
+	case AF_INET6:
+		addr6 = (struct sockaddr_in6 *)&name; 
+		addr6->sin6_family = AF_INET6;
+		addr6->sin6_port = local_port;
+		/* XXX: What will happen to multicasts if this is nonzero? */
+		memcpy(&addr6->sin6_addr,
 		       &local_address6, 
-		       sizeof(addr->sin6_addr));
-		name_len = sizeof(*addr);
+		       sizeof(addr6->sin6_addr));
+#ifdef HAVE_SA_LEN
+		addr6->sin6_len = sizeof(*addr6);
+#endif
+		name_len = sizeof(*addr6);
 		domain = PF_INET6;
-	} else { 
-#else 
-	{
+		if ((info->flags & INTERFACE_STREAMS) == INTERFACE_UPSTREAM) {
+			*do_multicast = 0;
+		}
+		break;
 #endif /* DHCPv6 */
-		struct sockaddr_in *addr = (struct sockaddr_in *)&name; 
+
+	case AF_INET:
+	default:
+		addr = (struct sockaddr_in *)&name; 
 		addr->sin_family = AF_INET;
 		addr->sin_port = local_port;
 		memcpy(&addr->sin_addr,
 		       &local_address,
 		       sizeof(addr->sin_addr));
+#ifdef HAVE_SA_LEN
+		addr->sin_len = sizeof(*addr);
+#endif
 		name_len = sizeof(*addr);
 		domain = PF_INET;
+		break;
 	}
 
 	/* Make a socket... */
@@ -167,6 +197,24 @@ if_register_socket(struct interface_info *info, int family, int do_multicast) {
 		log_fatal("Can't set SO_BROADCAST option on dhcp socket: %m");
 	}
 
+#if defined(DHCPv6) && defined(SO_REUSEPORT)
+	/*
+	 * We only set SO_REUSEPORT on AF_INET6 sockets, so that multiple
+	 * daemons can bind to their own sockets and get data for their
+	 * respective interfaces.  This does not (and should not) affect
+	 * DHCPv4 sockets; we can't yet support BSD sockets well, much
+	 * less multiple sockets.
+	 */
+	if (local_family == AF_INET6) {
+		flag = 1;
+		if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT,
+			       (char *)&flag, sizeof(flag)) < 0) {
+			log_fatal("Can't set SO_REUSEPORT option on dhcp "
+				  "socket: %m");
+		}
+	}
+#endif
+
 	/* Bind the socket to this interface's IP address. */
 	if (bind(sock, (struct sockaddr *)&name, name_len) < 0) {
 		log_error("Can't bind to dhcp address: %m");
@@ -179,7 +227,7 @@ if_register_socket(struct interface_info *info, int family, int do_multicast) {
 
 #if defined(SO_BINDTODEVICE)
 	/* Bind this socket to this interface. */
-	if (info->ifp &&
+	if ((local_family != AF_INET6) && (info->ifp != NULL) &&
 	    setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE,
 			(char *)(info -> ifp), sizeof(*(info -> ifp))) < 0) {
 		log_fatal("setsockopt: SO_BINDTODEVICE: %m");
@@ -222,38 +270,19 @@ if_register_socket(struct interface_info *info, int family, int do_multicast) {
 		}
 #endif
 	}
-	
-	if ((family == AF_INET6) && do_multicast) {
-		struct ipv6_mreq mreq;
 
-		/*
-		 * Join the DHCPv6 multicast groups so we will receive
-		 * multicast messages.
-		 */
-		if (inet_pton(AF_INET6, All_DHCP_Relay_Agents_and_Servers,
-			      &mreq.ipv6mr_multiaddr) <= 0) {
-			log_fatal("inet_pton: unable to convert '%s'", 
-				  All_DHCP_Relay_Agents_and_Servers);
-		}
-		mreq.ipv6mr_interface = if_nametoindex(info->name);
-		if (setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, 
-			       &mreq, sizeof(mreq)) < 0) {
-			log_fatal("setsockopt: IPV6_JOIN_GROUP: %m");
-		}
-		if (inet_pton(AF_INET6, All_DHCP_Servers,
-			      &mreq.ipv6mr_multiaddr) <= 0) {
-			log_fatal("inet_pton: unable to convert '%s'", 
-				  All_DHCP_Servers);
-		}
-		mreq.ipv6mr_interface = if_nametoindex(info->name);
-		if (setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, 
-			       &mreq, sizeof(mreq)) < 0) {
-			log_fatal("setsockopt: IPV6_JOIN_GROUP: %m");
+	if ((family == AF_INET6) &&
+	    ((info->flags & INTERFACE_UPSTREAM) != 0)) {
+		int hop_limit = 32;
+		if (setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
+			       &hop_limit, sizeof(int)) < 0) {
+			log_fatal("setsockopt: IPV6_MULTICAST_HOPS: %m");
 		}
 	}
 #endif /* DHCPv6 */
 
-	if (strcmp(info->name, "fallback") != 0)
+	/* If this is a normal IPv4 address, get the hardware address. */
+	if ((local_family == AF_INET) && (strcmp(info->name, "fallback") != 0))
 		get_hw_addr(info->name, &info->hw_address);
 
 	return sock;
@@ -333,15 +362,82 @@ void if_deregister_receive (info)
 
 
 #ifdef DHCPv6 
+/*
+ * This function joins the interface to DHCPv6 multicast groups so we will
+ * receive multicast messages.
+ */
+static void
+if_register_multicast(struct interface_info *info) {
+	int sock = info->rfdesc;
+	struct ipv6_mreq mreq;
+
+	if (inet_pton(AF_INET6, All_DHCP_Relay_Agents_and_Servers,
+		      &mreq.ipv6mr_multiaddr) <= 0) {
+		log_fatal("inet_pton: unable to convert '%s'", 
+			  All_DHCP_Relay_Agents_and_Servers);
+	}
+	mreq.ipv6mr_interface = if_nametoindex(info->name);
+	if (setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, 
+		       &mreq, sizeof(mreq)) < 0) {
+		log_fatal("setsockopt: IPV6_JOIN_GROUP: %m");
+	}
+
+	/*
+	 * The relay agent code sets the streams so you know which way
+	 * is up and down.  But a relay agent shouldn't join to the
+	 * Server address, or else you get fun loops.  So up or down
+	 * doesn't matter, we're just using that config to sense this is
+	 * a relay agent.
+	 */
+	if ((info->flags & INTERFACE_STREAMS) == 0) {
+		if (inet_pton(AF_INET6, All_DHCP_Servers,
+			      &mreq.ipv6mr_multiaddr) <= 0) {
+			log_fatal("inet_pton: unable to convert '%s'", 
+				  All_DHCP_Servers);
+		}
+		mreq.ipv6mr_interface = if_nametoindex(info->name);
+		if (setsockopt(sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, 
+			       &mreq, sizeof(mreq)) < 0) {
+			log_fatal("setsockopt: IPV6_JOIN_GROUP: %m");
+		}
+	}
+}
+
 void
 if_register6(struct interface_info *info, int do_multicast) {
-	info->rfdesc = if_register_socket(info, AF_INET6, do_multicast);
-	info->wfdesc = info->rfdesc;
+	/* Bounce do_multicast to a stack variable because we may change it. */
+	int req_multi = do_multicast;
+
+	if (global_v6_socket_references == 0) {
+		global_v6_socket = if_register_socket(info, AF_INET6,
+						      &req_multi);
+		if (global_v6_socket < 0) {
+			/*
+			 * if_register_socket() fatally logs if it fails to
+			 * create a socket, this is just a sanity check.
+			 */
+			log_fatal("Impossible condition at %s:%d", MDL);
+		} else {
+			log_info("Bound to *:%d", ntohs(local_port));
+		}
+	}
+		
+	info->rfdesc = global_v6_socket;
+	info->wfdesc = global_v6_socket;
+	global_v6_socket_references++;
+
+	if (req_multi)
+		if_register_multicast(info);
+
+	get_hw_addr(info->name, &info->hw_address);
+
 	if (!quiet_interface_discovery) {
 		if (info->shared_network != NULL) {
-			log_info("Listening on Socket/%s/%s", info->name, 
+			log_info("Listening on Socket/%d/%s/%s",
+				 global_v6_socket, info->name, 
 				 info->shared_network->name);
-			log_info("Sending on   Socket/%s/%s", info->name,
+			log_info("Sending on   Socket/%d/%s/%s",
+				 global_v6_socket, info->name,
 				 info->shared_network->name);
 		} else {
 			log_info("Listening on Socket/%s", info->name);
@@ -352,14 +448,16 @@ if_register6(struct interface_info *info, int do_multicast) {
 
 void 
 if_deregister6(struct interface_info *info) {
-	/* 
-	 * XXX: it would be nice to check for >= 0, but we need to change 
-	 *      interface_allocate() to set the file descriptors for that.
-	 */
-	close(info->rfdesc);	
-	info->rfdesc = -1;
-	close(info->wfdesc);	
-	info->wfdesc = -1;
+	/* Dereference the global v6 socket. */
+	if ((info->rfdesc == global_v6_socket) &&
+	    (info->wfdesc == global_v6_socket) &&
+	    (global_v6_socket_references > 0)) {
+		global_v6_socket_references--;
+		info->rfdesc = -1;
+		info->wfdesc = -1;
+	} else {
+		log_fatal("Impossible condition at %s:%d", MDL);
+	}
 
 	if (!quiet_interface_discovery) {
 		if (info->shared_network != NULL) {
@@ -372,6 +470,13 @@ if_deregister6(struct interface_info *info) {
 			log_info("Disabling output on Socket/%s", info->name);
 		}
 	}
+
+	if (global_v6_socket_references == 0) {
+		close(global_v6_socket);
+		global_v6_socket = -1;
+
+		log_info("Unbound from *:%d", ntohs(local_port));
+	}
 }
 #endif /* DHCPv6 */
 
@@ -461,6 +566,21 @@ static size_t CMSG_SPACE(size_t len) {
 #endif /* DHCPv6 */
 
 #ifdef DHCPv6
+/*
+ * For both send_packet6() and receive_packet6() we need to allocate
+ * space for the cmsg header information.  We do this once and reuse
+ * the buffer.
+ */
+static void   *control_buf = NULL;
+static size_t  control_buf_len = 0;
+
+static void
+allocate_cmsg_cbuf(void) {
+	control_buf_len = CMSG_SPACE(sizeof(struct in6_pktinfo));
+	control_buf = dmalloc(control_buf_len, MDL);
+	return;
+}
+
 /* 
  * For both send_packet6() and receive_packet6() we need to use the 
  * sendmsg()/recvmsg() functions rather than the simpler send()/recv()
@@ -488,10 +608,20 @@ ssize_t send_packet6(struct interface_info *interface,
 	int result;
 	struct in6_pktinfo *pktinfo;
 	struct cmsghdr *cmsg;
-	union {
-		struct cmsghdr cmsg_sizer;
-		u_int8_t pktinfo_sizer[CMSG_SPACE(sizeof(struct in6_pktinfo))];
-	} control_buf;
+
+	/*
+	 * If necessary allocate space for the control message header.
+	 * The space is common between send and receive.
+	 */
+
+	if (control_buf == NULL) {
+		allocate_cmsg_cbuf();
+		if (control_buf == NULL) {
+			log_error("send_packet6: unable to allocate cmsg header");
+			return(ENOMEM);
+		}
+	}
+	memset(control_buf, 0, control_buf_len);
 
 	/*
 	 * Initialize our message header structure.
@@ -522,8 +652,8 @@ ssize_t send_packet6(struct interface_info *interface,
 	 * source address if we wanted, but we can safely let the
 	 * kernel decide what that should be. 
 	 */
-	m.msg_control = &control_buf;
-	m.msg_controllen = sizeof(control_buf);
+	m.msg_control = control_buf;
+	m.msg_controllen = control_buf_len;
 	cmsg = CMSG_FIRSTHDR(&m);
 	cmsg->cmsg_level = IPPROTO_IPV6;
 	cmsg->cmsg_type = IPV6_PKTINFO;
@@ -580,17 +710,29 @@ ssize_t receive_packet (interface, buf, len, from, hfrom)
 ssize_t 
 receive_packet6(struct interface_info *interface, 
 		unsigned char *buf, size_t len, 
-		struct sockaddr_in6 *from, struct in6_addr *to_addr) {
+		struct sockaddr_in6 *from, struct in6_addr *to_addr,
+		unsigned int *if_idx)
+{
 	struct msghdr m;
 	struct iovec v;
 	int result;
 	struct cmsghdr *cmsg;
 	struct in6_pktinfo *pktinfo;
-	int found_to_addr;
-	union {
-	        struct cmsghdr cmsg_sizer;
-	        u_int8_t pktinfo_sizer[CMSG_SPACE(sizeof(struct in6_pktinfo))];
-	} control_buf;
+	int found_pktinfo;
+
+	/*
+	 * If necessary allocate space for the control message header.
+	 * The space is common between send and receive.
+	 */
+	if (control_buf == NULL) {
+		allocate_cmsg_cbuf();
+		if (control_buf == NULL) {
+			log_error("receive_packet6: unable to allocate cmsg "
+				  "header");
+			return(ENOMEM);
+		}
+	}
+	memset(control_buf, 0, control_buf_len);
 
 	/*
 	 * Initialize our message header structure.
@@ -621,8 +763,8 @@ receive_packet6(struct interface_info *interface,
 	 * information (when we initialized the interface), so we
 	 * should get the destination address from that.
 	 */
-	m.msg_control = &control_buf;
-	m.msg_controllen = sizeof(control_buf);
+	m.msg_control = control_buf;
+	m.msg_controllen = control_buf_len;
 
 	result = recvmsg(interface->rfdesc, &m, 0);
 
@@ -635,18 +777,19 @@ receive_packet6(struct interface_info *interface,
 		 * We also keep a flag to see if we found it. If we 
 		 * didn't, then we consider this to be an error.
 		 */
-		found_to_addr = 0;
+		found_pktinfo = 0;
 		cmsg = CMSG_FIRSTHDR(&m);
 		while (cmsg != NULL) {
 			if ((cmsg->cmsg_level == IPPROTO_IPV6) && 
 			    (cmsg->cmsg_type == IPV6_PKTINFO)) {
 				pktinfo = (struct in6_pktinfo *)CMSG_DATA(cmsg);
 				*to_addr = pktinfo->ipi6_addr;
-				found_to_addr = 1;
+				*if_idx = pktinfo->ipi6_ifindex;
+				found_pktinfo = 1;
 			}
 			cmsg = CMSG_NXTHDR(&m, cmsg);
 		}
-		if (!found_to_addr) {
+		if (!found_pktinfo) {
 			result = -1;
 			errno = EIO;
 		}
@@ -669,7 +812,7 @@ isc_result_t fallback_discard (object)
 	struct interface_info *interface;
 
 	if (object -> type != dhcp_type_interface)
-		return ISC_R_INVALIDARG;
+		return DHCP_R_INVALIDARG;
 	interface = (struct interface_info *)object;
 
 	status = recvfrom (interface -> wfdesc, buf, sizeof buf, 0,
